#include "config.h"

program ccs_qcd_solver_bench
!***********************************************************************
!$Id: ccs_qcd_solver_bench.F90,v 1.4 2012/12/19 02:17:33 ishikawa Exp $
!
!  CCS QCD solver benchmark test program.
!
!  This benchmark test program measures CPU and MPI performances.
!
!***********************************************************************
  use comlib
  use lattice_class
  use ccs_qcd_solver_bench_class
  implicit none
!***********************************************************************
  complex(8) ::  ye(COL,SPIN,0:NTH,0:NZ1,0:NY1,0:NX1)
  complex(8) ::  yo(COL,SPIN,0:NTH,0:NZ1,0:NY1,0:NX1)
  complex(8) ::  te(COL,SPIN,0:NTH,0:NZ1,0:NY1,0:NX1)
  complex(8) :: yde(COL,SPIN,0:NTH,0:NZ1,0:NY1,0:NX1)
  complex(8) :: ydo(COL,SPIN,0:NTH,0:NZ1,0:NY1,0:NX1)
  complex(8) ::  ue(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)
  complex(8) ::  uo(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)
  complex(8) :: fclinve(CLSPH,0:NTH,NZ,NY,NX,2)
  complex(8) :: fclinvo(CLSPH,0:NTH,NZ,NY,NX,2)
  complex(8), allocatable :: ucle(:,:,:,:,:,:,:)
  complex(8), allocatable :: uclo(:,:,:,:,:,:,:)
  real(8) :: bicg_dp_flop, tclv_flop, flop
  real(8) :: kappa,csw,tol,logdetfcl
  integer :: iiter
  integer :: ix,iy,iz,itb,ieoxyz,ic,is
  real(8) :: rnorm1,rnorm2,rnorm3
  integer :: iargc
  character(256) :: arg(6)
  integer :: iloop,NLOOP
  real(8) :: etime0,etime1,etime2
  real(8) :: bicg_dp_etime,bicg_total_etime,clv_etime
  integer, parameter :: iout=_FILE_IO_BICGSTAB_HMC


  call initset

  mult_iter=0
  bicg_total_etime=0.0d0
  bicg_dp_etime=0.0d0
  clv_etime=0.0d0
  copygtime=0.0d0
  copyftime=0.0d0

  if ( COMMAND_ARGUMENT_COUNT() /= 4) then
!  if (iargc() /= 4) then
    if (nodeid==0) then
      write(*,*)
      write(*,'("Usage: ccs_qcd_solver_bench loop_count kappa csw tolerance")')
      write(*,*)
      write(*,'("CCS QCD Solver Benchmark Program:")')
      write(*,'("Solve QCD sparse linear equiation, A x = b, using BiCGStab algorithm, where")')
      write(*,'("  A : Clover-term red-black preconditioned Wilson-Dirac QCD matrix given by")')
      write(*,'("  A = 1ee - kappa^2 Fee^-1 Meo Foo^-1 Moe, where")')
      write(*,'("    Moe (Meo)       : Hopping matrix from even (odd) sites to odd (even) sites,")')
      write(*,'("    Fee^-1 (Foo^-1) : Inverse clover term on even (odd) sites, and")')
      write(*,'("    x, b            : Quark field vectors on even sites.")')
      write(*,*)
      write(*,'(" Command args:")')
      write(*,'(" loop_count       Execution count for BiCGStab solver. (integer number)")')
      write(*,'(" kappa            Hopping parameter, should be near 0.125d0. (double prec. number)")')
      write(*,'(" csw              Clover coefficient, should be near 1.0d0. (double prec. number)")')
      write(*,'(" tolerance        Tolerance for BiCGStab solver. for ex. 1.0d-14. (double prec. number)")')
      write(*,'(80("="))')
      write(*,'(" CCS QCD Solver Benchmark Program")')
      write(*,'(X,A)')_VERSION_
      write(*,'(X,A)')_REVISION_
      write(*,'(80("="))')
    endif
#ifndef _singlePU
    call comlib_finalize
#endif
    stop
  endif
  if (nodeid==0) then
    call  GET_COMMAND_ARGUMENT(1,arg(1))
    call  GET_COMMAND_ARGUMENT(2,arg(2))
    call  GET_COMMAND_ARGUMENT(3,arg(3))
    call  GET_COMMAND_ARGUMENT(4,arg(4))
!    call getarg(1,arg(1))
!    call getarg(2,arg(2))
!    call getarg(3,arg(3))
!    call getarg(4,arg(4))
    read(arg(1),*)NLOOP
    read(arg(2),*)kappa
    read(arg(3),*)csw
    read(arg(4),*)tol
  endif
#ifndef _singlePU
  call comlib_bcast(NLOOP,0)
  call comlib_bcast(kappa,0)
  call comlib_bcast(csw,0)
  call comlib_bcast(tol,0)
#endif


  if (nodeid==0) then
    open(iout,file="solver_residual.log",form="formatted")
    write(*,'(80("="))')
    write(*,'(" CCS QCD Solver Benchmark Program")')
    write(*,'(X,A)')_VERSION_
    write(*,'("    Lattice Size :",3I3,I4)')NTX,NTY,NTZ,NTT
    write(*,'(" Node Array Size :",3I3)')NDIMX,NDIMY,NDIMZ
    write(*,'(" PU Lattice Size :",3I3,I4,"*2")')NX,NY,NZ,NTH
    write(*,'("            Loop :",I9)')NLOOP
    write(*,'("           Kappa :",F14.6)')kappa
    write(*,'("             CSW :",F14.6)')csw
    write(*,'("             Tol :",E14.6)')tol
    write(*,'(80("="))')
  endif

!***************************
! initialize
! (ue,uo) and (ye,yo)
! with Gaussian rand. num.
!***************************
  call init_u_and_y(ue,uo,ye,yo)

!***************************
! calculate clover term
!***************************
  call xclock(etime0,8)
  allocate(ucle(COL,COL,NTH,NZ,NY,NX,NDIM*(NDIM-1)/2))
  allocate(uclo(COL,COL,NTH,NZ,NY,NX,NDIM*(NDIM-1)/2))
  call clover(ucle,uclo,ue,uo)
  call clvinv(kappa,csw,fclinve,fclinvo,ucle,uclo,logdetfcl,1)
  deallocate(ucle,uclo)
  call xclock(etime1,8)
  clv_etime=clv_etime+(etime1-etime0)
  if (nodeid==0) then
    write(*,*)
    write(*,*)" Sum[Log[Det[F[n]]],n] = ",logdetfcl
    write(*,*)
  endif
!  tclv_flop = 18.d0*(880.d0 + 93.d0*(-1 + NDIM)*NDIM)*NTH*NX*NY*NZ
  tclv_flop = flop_count_clover + flop_count_clvinv

  call output(ue,uo,ye,yo)

!************************
! BiCGStab test
!************************
  bicg_dp_flop = 0.0d0
  do iloop=1,NLOOP
    iiter=0
!************************
! solve A x = b
! ye  <= b (input)
! yde => x (output)
!************************
    call xclock(etime0,8)
    call assign_inv_mult_wd(tol,iiter,flop,etime2,  &
 &                      kappa,ye,yo,yde,ydo,ue,uo,0, &
 &                                          fclinve,fclinvo)
   
!    call bicgstab_hmc(tol,iiter,flop,etime2,  &
! &                    kappa,ye,yde,ue,uo,0,fclinve,fclinvo)
    bicg_dp_flop = bicg_dp_flop + flop
    bicg_dp_etime = bicg_dp_etime + etime2
    call xclock(etime1,8)
    bicg_total_etime = bicg_total_etime + (etime1-etime0)
    call output(ue,uo,ye,yo)

    rnorm3=0.0d0
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz,2)
      do itb=1-ieoxyz,NTH-ieoxyz
        do is=1,SPIN
        do ic=1,COL
          rnorm3=rnorm3+ real(yde(ic,is,itb,iz,iy,ix))**2  &
 &                     +aimag(yde(ic,is,itb,iz,iy,ix))**2
        enddo
        enddo
      enddo
      do itb=ieoxyz,NTH+ieoxyz-1
        do is=1,SPIN
        do ic=1,COL
          rnorm3=rnorm3+ real(ydo(ic,is,itb,iz,iy,ix))**2  &
 &                     +aimag(ydo(ic,is,itb,iz,iy,ix))**2
        enddo
        enddo
      enddo
    enddo
    enddo
    enddo

#ifndef _singlePU
    call comlib_sumcast(rnorm3)
#endif

    if (nodeid==0) then
      write(*,'("  Iter = ",I5)')iiter
      write(*,'("|D\y|^2= ",E24.15)')rnorm3
    endif
  enddo

!******************
! write results
!******************
  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'("  Clover + Clover_inv Performance.")')
    write(*,'("           Flop [MFlop/node] :",32(F12.4))') tclv_flop*1.d-6
    write(*,'("          Elapsed Time [sec] :",32(F12.4))') clv_etime
    write(*,'("Performance [MFlop/sec/node] :",32(F12.4))') tclv_flop*1.d-6/clv_etime
    write(*,'(80("-"))')
    write(*,'("  BiCGStab(CPU:double precision) Performance.")')
    write(*,'("           Flop [MFlop/node] :",32(F12.4))') bicg_dp_flop*1.d-6
    write(*,'("          Elapsed Time [sec] :",32(F12.4))') bicg_dp_etime
    write(*,'("Performance [MFlop/sec/node] :",32(F12.4))') bicg_dp_flop*1.d-6/bicg_dp_etime
    write(*,'("      Full Copy_y Time [sec] :",32(F12.4))') copyftime
#if _NDIMX != 1
    write(*,'("    MPI Commun. SizeX[Mbyte] :",32(F12.4))') dble(copy_y_size(1)*16)/1024/1024
    write(*,'("    MPI Commun. TimeX  [sec] :",32(F12.4))') copy_y_time(1)
    write(*,'("    MPI Commun. BandX [MB/s] :",32(F12.4))') dble(copy_y_size(1)*16)/1024/1024/copy_y_time(1)
#endif
#if _NDIMY != 1
    write(*,'("    MPI Commun. SizeY[Mbyte] :",32(F12.4))') dble(copy_y_size(2)*16)/1024/1024
    write(*,'("    MPI Commun. TimeY  [sec] :",32(F12.4))') copy_y_time(2)
    write(*,'("    MPI Commun. BandY [MB/s] :",32(F12.4))') dble(copy_y_size(2)*16)/1024/1024/copy_y_time(2)
#endif
#if _NDIMZ != 1
    write(*,'("    MPI Commun. SizeZ[Mbyte] :",32(F12.4))') dble(copy_y_size(3)*16)/1024/1024
    write(*,'("    MPI Commun. TimeZ  [sec] :",32(F12.4))') copy_y_time(3)
    write(*,'("    MPI Commun. BandZ [MB/s] :",32(F12.4))') dble(copy_y_size(3)*16)/1024/1024/copy_y_time(3)
#endif
    write(*,'(80("-"))')
    write(*,'("  BiCGStab Total")')
    write(*,'("          Elapsed Time [sec] :",32(F12.4))') bicg_total_etime
    write(*,'(80("="))')
    close(iout)
  endif

#ifndef _singlePU
  call comlib_finalize
#endif

  stop
end program
